/*
  MipSmooth  (c) 2003 by Klaus Post

	This program is free software; you can redistribute it and/or modify
	it under the terms of the GNU General Public License as published by
	the Free Software Foundation.

	This program is distributed in the hope that it will be useful,
	but WITHOUT ANY WARRANTY; without even the implied warranty of
	MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
	GNU General Public License for more details.

	You should have received a copy of the GNU General Public License
	along with this program; if not, write to the Free Software
	Foundation, Inc., 675 Mass Ave, Cambridge, MA 02139, USA.

	The author can be contacted at:
	sh0dan@stofanet.dk
*/

#include <crtdbg.h>
#include <windows.h>
#include "avisynth.h"
#include "info.h"
#include <stdio.h>
#include <math.h>

#define MAX_MIPS 4

const HKEY RegRootKey = HKEY_LOCAL_MACHINE;
const char RegAvisynthKey[] = "Software\\Avisynth";


class MipSmooth : public GenericVideoFilter
{
  PClip *curr_mips;     // Array of mip clips.
  PVideoFrame *mips;     // Array of mip frames.
  PVideoFrame *work_mips;     // Array of mip frames.
  const BYTE** planeP;  // Plane pointers
  const BYTE** work_planeP;  // Plane pointers
  int* planePitch;      // Plane pitches (corresponds to planeP)
  int* work_planePitch;      // Plane pitches (corresponds to work_planeP)
  BYTE* multipliers;    // Multipliers for each plane (8 bytes per plane)
  BYTE* thresholds;    // Thresholds for each plane (8 bytes per plane)
  double scenechange;
  int spat_thresh;
  int temp_thresh;
  int spat_threshUV;
  int temp_threshUV;
  int method;
  bool show;
  bool weigh;
  bool supertwo;
  double scale_factor;
  double scenechange_factor;

  enum { 
    METHOD_SOFT=1,
    METHOD_STRONG=2,
    METHOD_SUPERSOFT=3,
    METHOD_SUPERSTRONG=4
  };


public:
    MipSmooth(PClip _child, int sp_t, int te_t, int sp_tUV, int te_tUV, double sc, const char* _method, const char* downsizer, const char* upsizer, double factor, bool _show, bool _weigh, IScriptEnvironment* env) :
	GenericVideoFilter(_child), spat_thresh(sp_t),temp_thresh(te_t), scenechange(sc), spat_threshUV(sp_tUV),temp_threshUV(te_tUV), show(_show), weigh(_weigh)
	{

    supertwo = false;

    curr_mips = new PClip[MAX_MIPS];

    if (!(env->GetCPUFlags() & CPUF_INTEGER_SSE))
      env->ThrowError("MipSmooth: Requires Integer SSE capable CPU. (Athlon, P3 or newer)");

    if (vi.width & 3)
      env->ThrowError("MipSmooth: Mod 4 width resolution is required.");

    if (vi.IsYUY2() && (vi.width&7))
      env->ThrowError("MipSmooth: Mod 8 width resolution is required in YUY2 mode.");
    
    if (lstrcmpi(_method, "soft") == 0)
      method = METHOD_SOFT;
    else if (lstrcmpi(_method, "strong") == 0)
      method = METHOD_STRONG;
    else if (lstrcmpi(_method, "superstrong") == 0)
      method = METHOD_SUPERSTRONG;
    else if (lstrcmpi(_method, "supersoft") == 0)
      method = METHOD_SUPERSOFT;
    else if (lstrcmpi(_method, "superstrong2") == 0) {
      method = METHOD_SUPERSTRONG;
      supertwo = true;
    } else if (lstrcmpi(_method, "supersoft2") == 0) {
      method = METHOD_SUPERSOFT;
      supertwo = true;
    } else 
      env->ThrowError("MipSmooth: Method must be either \"soft\" or \"strong\"!");

    if (method<3)
      weigh=false;


    // Remove reduce as an option - shifts chroma
    if (lstrcmpi(downsizer, "reduce") == 0) {
      downsizer = "bilinear";
    }

    scenechange=min(255.0,max(scenechange,0.0));
    temp_thresh=min(255,max(temp_thresh,0));
    temp_threshUV=min(255,max(temp_threshUV,0));
    spat_thresh=min(255,max(spat_thresh,0));
    spat_threshUV=min(255,max(spat_threshUV,0));

    const char* upsizerString;

    if (lstrcmpi(upsizer, "bilinear") == 0) 
      upsizerString="BilinearResize";
    else if (lstrcmpi(upsizer, "bicubic") == 0)
      upsizerString="BicubicResize";
    else if (lstrcmpi(upsizer, "lanczos") == 0)
      upsizerString="LanczosResize";
    else if (lstrcmpi(upsizer, "point") == 0)
      upsizerString="PointResize";
    else 
      env->ThrowError("MipSmooth: Upsizer must be either \"Point\", \"Bilinear\", \"Bicubic\" or \"Lanczos\"");

    const char* downsizerString;
    if (lstrcmpi(downsizer, "bilinear") == 0) 
      downsizerString="BilinearResize";
    else if (lstrcmpi(downsizer, "bicubic") == 0)
      downsizerString="BicubicResize";
    else if (lstrcmpi(downsizer, "lanczos") == 0)
      downsizerString="LanczosResize";
    else if (lstrcmpi(downsizer, "point") == 0)
      downsizerString="LanczosResize";
    else 
      env->ThrowError("MipSmooth: Downsizer must be either \"Point\", \"Reduce\", \"Bilinear\", \"Bicubic\" or \"Lanczos\"");
    
    AVSValue Targs[3] = { child, max(16,((int)((double)vi.width*factor)/4)*4), max(16,((int)((double)vi.height*factor)/4)*4) }; 
    
    PClip reduced = env->Invoke(downsizerString,AVSValue(Targs,3)).AsClip(); 
    Targs[0] = reduced;
    
    try {
      reduced = env->Invoke("InternalCache",AVSValue(Targs,1)).AsClip();
      reduced->SetCacheHints(CACHE_RANGE,1);
    } catch (IScriptEnvironment::NotFound) {  }
      
    AVSValue up_args[3] = { reduced, vi.width, vi.height }; 
    curr_mips[0] = env->Invoke(upsizerString, AVSValue(up_args,3)).AsClip();
    
    Targs[0] = reduced;
    Targs[1] = max(16, ((int)(Targs[1].AsFloat()*factor)/4)*4);
    Targs[2] = max(16, ((int)(Targs[2].AsFloat()*factor)/4)*4);
    
    reduced = env->Invoke(downsizerString,AVSValue(Targs,3)).AsClip(); 
    Targs[0] = reduced;
    
    try {
      reduced = env->Invoke("InternalCache",AVSValue(Targs,1)).AsClip();
      reduced->SetCacheHints(CACHE_RANGE,1);
     } catch (IScriptEnvironment::NotFound) {  }
      
     up_args[0] = reduced;
     curr_mips[1] = env->Invoke(upsizerString,AVSValue(up_args,3)).AsClip();
      
     Targs[0] = reduced;
     Targs[1] = max(16, ((int)(Targs[1].AsFloat()*factor)/4)*4);
     Targs[2] = max(16, ((int)(Targs[2].AsFloat()*factor)/4)*4);
      
     reduced = env->Invoke(downsizerString,AVSValue(Targs,3)).AsClip(); 
     up_args[0] = reduced;
     curr_mips[2] = env->Invoke(upsizerString,AVSValue(up_args,3)).AsClip();


    AVSValue args[1] = { child }; 

    for (int i = 0; i < 3; i++) {
      args[0] = curr_mips[i];
      try {
        curr_mips[i] = env->Invoke("InternalCache",AVSValue(args,1)).AsClip();
        curr_mips[i]->SetCacheHints(CACHE_RANGE,1);
      } catch (IScriptEnvironment::NotFound) {  
        // ignore - we can live without the cache.
      }
    }

    mips = new PVideoFrame[MAX_MIPS*3];
    work_mips = new PVideoFrame[MAX_MIPS];
    planeP = new const BYTE*[MAX_MIPS*3];
    planePitch = new int[MAX_MIPS*3];
    thresholds = new BYTE[8*MAX_MIPS*3];

    scenechange_factor = 1.0;

    scenechange *= (double)((vi.width/32)*32)*vi.height;
    if (vi.IsRGB()) scenechange_factor = 3.0;
    child->SetCacheHints(CACHE_RANGE,1);

    scale_factor = factor;
	}

  MipSmooth::~MipSmooth(void) {
    delete[] curr_mips;
    delete[] mips;
    delete[] planeP;
    delete[] planePitch;
    delete[] thresholds;
    delete[] work_mips;
  }
    PVideoFrame __stdcall GetFrame(int n, IScriptEnvironment* env);
    void isse_accumulate_line_org(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, int t, int div);
    void isse_accumulate_line_org_soft(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, int t, int div);
    void isse_accumulate_line_org_weigh(const BYTE* c_plane, const BYTE** planeP,  int rowsize, int t, int div);
    void isse_accumulate_line_org_soft_weigh(const BYTE* c_plane, const BYTE** planeP,  int rowsize, int t, int div);
    void isse_accumulate_line_strong(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, BYTE* t, int div);
    void isse_accumulate_line_soft(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, BYTE* t, int div);
    void isse_blendline_soft(const BYTE* c_plane, const BYTE** planeP, int planes, int rowsize, int div);
    int isse_scenechange(const BYTE* c_plane, const BYTE* tplane, int height, int width, int c_pitch, int t_pitch, int pixel_type);
};

#include "mip_asm.cpp"

PVideoFrame __stdcall MipSmooth::GetFrame(int n, IScriptEnvironment* env)
{
  int cmip=0;
  int cPlane = PLANAR_Y;
	char buf[80];

  BYTE* t_th = new BYTE[MAX_MIPS*3];
  BYTE* t_thUV = new BYTE[MAX_MIPS*3];
  PVideoFrame src = child->GetFrame(n,env);

  env->MakeWritable(&src);
  BYTE* srcp = src->GetWritePtr(cPlane);
  int w = src->GetRowSize(cPlane);
  int h = src->GetHeight(cPlane);
  int src_pitch = src->GetPitch(cPlane);
  int sc_prev, sc_next;
  int supermips = 0;

  if (method==METHOD_SUPERSOFT || method==METHOD_SUPERSTRONG) {

    work_planeP = new const BYTE*[MAX_MIPS];
    work_planePitch = new int[MAX_MIPS];
    int* w_cPlane = new int[4];

    if (vi.IsPlanar()) {
      w_cPlane[0] = PLANAR_Y;
      w_cPlane[1] = PLANAR_U;
      w_cPlane[2] = PLANAR_V;
      w_cPlane[3] = -1;
    } else {
      w_cPlane[0] = PLANAR_Y;
      w_cPlane[1] = -1;
    }

    // Temporal frame -1 Supermip
    mips[cmip] = child->GetFrame(n-1,env); 
    sc_prev = isse_scenechange(srcp, mips[cmip]->GetReadPtr(cPlane), h,w,src_pitch, mips[cmip]->GetPitch(cPlane),vi.pixel_type);
    mips[cmip] = child->GetFrame(n,env); 
    if (sc_prev<scenechange * scenechange_factor) {
      env->MakeWritable(&mips[cmip]);
      // Arrange for previous frame supermip
      int w_mip=0;
      work_mips[w_mip] = curr_mips[0]->GetFrame(n-1,env);
      w_mip++;
      work_mips[w_mip] = curr_mips[1]->GetFrame(n-1,env);
      w_mip++;
      work_mips[w_mip] = curr_mips[2]->GetFrame(n-1,env);
      w_mip++;
      int nP=0;
      while (w_cPlane[nP]>0) {

        for (int i = 0;i<w_mip; i++) {
          work_planeP[i] = work_mips[i]->GetReadPtr(w_cPlane[nP]);
          work_planePitch[i] = work_mips[i]->GetPitch(w_cPlane[nP]);
        }

        int w_thresh = (w_cPlane[nP] == PLANAR_Y) ? spat_thresh : spat_threshUV;
        w_thresh = w_thresh|(w_thresh<<8)|(w_thresh<<16)|(w_thresh<<24);
        int c_div;
        if (method == METHOD_SUPERSTRONG) {
          c_div = (weigh) ?32767/(5+w_mip-1) : 32767/(w_mip-1);
        } else {
          c_div = (weigh) ?  32767/(8+w_mip) : 32767/(w_mip);
        }
        BYTE* work_src = mips[cmip]->GetWritePtr(w_cPlane[nP]);
        int w_h = mips[cmip]->GetHeight(w_cPlane[nP]);
        if (w_thresh) {
          for (int y=0;y<w_h;y++) {
            if (method == METHOD_SUPERSTRONG) {
              if (weigh)
                  isse_accumulate_line_org_weigh(work_src, work_planeP, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div/4);
                else
                  isse_accumulate_line_org(work_src, work_planeP, w_mip-1, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
            } else {
            if (weigh)
              isse_accumulate_line_org_soft_weigh(work_src, work_planeP, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
             else
              isse_accumulate_line_org_soft(work_src, work_planeP, w_mip-1, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
            }
            work_src+=mips[cmip]->GetPitch(w_cPlane[nP]);
            for (int i=0 ;i<w_mip;i++) {
              work_planeP[i] += work_planePitch[i];
            }
          }
        }

        nP++;
      } // End while    

      t_th[cmip] = (temp_thresh); t_thUV[cmip] = (temp_threshUV);
      cmip++;
      supermips += 3;
    } // end if scenechange


    // Current frame  Supermip
    mips[cmip] = child->GetFrame(n,env); 
    env->MakeWritable(&mips[cmip]);

    // Arrange for current frame supermip
    int w_mip=0;
    work_mips[w_mip] = curr_mips[0]->GetFrame(n,env);
    w_mip++;
    work_mips[w_mip] = curr_mips[1]->GetFrame(n,env);
    w_mip++;
    work_mips[w_mip] = curr_mips[2]->GetFrame(n,env);
    w_mip++;
    int nP=0;

    while (w_cPlane[nP]>0) {
      for (int i = 0; i < w_mip; i++) {
        work_planeP[i] = work_mips[i]->GetReadPtr(w_cPlane[nP]);
        work_planePitch[i] = work_mips[i]->GetPitch(w_cPlane[nP]);
      }

      int w_thresh = (w_cPlane[nP]==PLANAR_Y) ? spat_thresh : spat_threshUV;
      w_thresh = w_thresh|(w_thresh<<8)|(w_thresh<<16)|(w_thresh<<24);
        int c_div;
        if (method == METHOD_SUPERSTRONG) {
          c_div = (weigh) ? 32767/(5+w_mip-1) : 32767/(w_mip-1);
        } else {
          c_div = (weigh) ?  32767/(8+w_mip) : 32767/(w_mip);
        }
      BYTE* work_src = mips[cmip]->GetWritePtr(w_cPlane[nP]);
      int w_h = mips[cmip]->GetHeight(w_cPlane[nP]);

      if (w_thresh) {
        for (int y=0;y<w_h;y++) {
          if (method == METHOD_SUPERSTRONG) {
            if (weigh)
              isse_accumulate_line_org_weigh(work_src, work_planeP, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
             else
              isse_accumulate_line_org(work_src, work_planeP, w_mip-1, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
          } else {
            if (weigh)
              isse_accumulate_line_org_soft_weigh(work_src, work_planeP, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
             else
              isse_accumulate_line_org_soft(work_src, work_planeP, w_mip-1, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
          }
          work_src += mips[cmip]->GetPitch(w_cPlane[nP]);
          for (int i=0 ; i < w_mip ; i++) {
            work_planeP[i] += work_planePitch[i];
          }
        }
      }

      nP++;
    } // End while    
    t_th[cmip] = spat_thresh; t_thUV[cmip] = spat_threshUV;
    cmip++;
    supermips += 3;

    // Next frame  Supermip
    mips[cmip] = child->GetFrame(n+1,env); 

    sc_next = isse_scenechange(srcp, mips[cmip]->GetReadPtr(cPlane), h,w,src_pitch, mips[cmip]->GetPitch(cPlane),vi.pixel_type);

    mips[cmip] = child->GetFrame(n,env); 

    if (sc_next < scenechange * scenechange_factor) {

      env->MakeWritable(&mips[cmip]);

      // Arrange for Next frame supermip
      int w_mip=0;
      work_mips[w_mip] = curr_mips[0]->GetFrame(n+1,env);
      w_mip++;
      work_mips[w_mip] = curr_mips[1]->GetFrame(n+1,env);
      w_mip++;
      work_mips[w_mip] = curr_mips[2]->GetFrame(n+1,env);
      w_mip++;

      int nP=0;
      // Create supermip
      while (w_cPlane[nP]>0) {
        for (int i = 0;i<w_mip; i++) {
          work_planeP[i] = work_mips[i]->GetReadPtr(w_cPlane[nP]);
          work_planePitch[i] = work_mips[i]->GetPitch(w_cPlane[nP]);
        }

        int w_thresh = (w_cPlane[nP]==PLANAR_Y) ? spat_thresh : spat_threshUV;
        w_thresh = w_thresh|(w_thresh<<8)|(w_thresh<<16)|(w_thresh<<24);
        int c_div;
        if (method == METHOD_SUPERSTRONG) {
          c_div = (weigh) ?32767/(5+w_mip-1) : 32767/(w_mip-1);
        } else {
          c_div = (weigh) ?  32767/(8+w_mip) : 32767/(w_mip);
        }
        BYTE* work_src = mips[cmip]->GetWritePtr(w_cPlane[nP]);
        int w_h = mips[cmip]->GetHeight(w_cPlane[nP]);

        if (w_thresh) {
          for (int y=0; y < w_h ; y++) {
            if (method == METHOD_SUPERSTRONG) {
              if (weigh)
                isse_accumulate_line_org_weigh(work_src, work_planeP, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
              else
                isse_accumulate_line_org(work_src, work_planeP, w_mip-1, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
            } else {
            if (weigh)
              isse_accumulate_line_org_soft_weigh(work_src, work_planeP, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
             else
              isse_accumulate_line_org_soft(work_src, work_planeP, w_mip-1, mips[cmip]->GetRowSize(w_cPlane[nP]), w_thresh, c_div);
            }
            work_src+=mips[cmip]->GetPitch(w_cPlane[nP]);
            for (int i=0 ;i<w_mip;i++) {
              work_planeP[i] += work_planePitch[i];
            }
          }
        }
        nP++;
      } // End while    

      t_th[cmip] = (temp_thresh); t_thUV[cmip] = (temp_threshUV);
      cmip++;
      supermips += 3;
    } // end if scenechange

    delete[] work_planeP;
    delete[] work_planePitch;
    delete[] w_cPlane;

  } else {   // STRONG or SOFT
    // Spatial current frame
    
    mips[cmip] = curr_mips[0]->GetFrame(n,env); t_th[cmip] = spat_thresh; t_thUV[cmip] = spat_threshUV;
    cmip++;
    mips[cmip] = curr_mips[1]->GetFrame(n,env); t_th[cmip] = (spat_thresh+1)/2; t_thUV[cmip] = (spat_threshUV+1)/2;
    cmip++;
    mips[cmip] = curr_mips[2]->GetFrame(n,env); t_th[cmip] = (spat_thresh+2)/3; t_thUV[cmip] = (spat_threshUV+1)/2;
    cmip++;
    
    // Temporal frame -1 
    
    mips[cmip] = child->GetFrame(n-1,env); 
    sc_prev = isse_scenechange(srcp, mips[cmip]->GetReadPtr(cPlane), h,w,src_pitch, mips[cmip]->GetPitch(cPlane),vi.pixel_type);
    if (sc_prev < scenechange * scenechange_factor) {
      t_th[cmip] = temp_thresh;
      t_thUV[cmip] = temp_threshUV;
      cmip++;
      mips[cmip] = curr_mips[0]->GetFrame(n-1,env); t_th[cmip] = (temp_thresh+1)/2; t_thUV[cmip] = (temp_threshUV+1)/2;
      cmip++;
      mips[cmip] = curr_mips[1]->GetFrame(n-1,env); t_th[cmip] = (temp_thresh+2)/3; t_thUV[cmip] = (temp_threshUV+1)/2;
      cmip++;
    }
    
    // Temporal frame +1 
    
    
    mips[cmip] = child->GetFrame(n+1,env);
    
    sc_next = isse_scenechange(srcp, mips[cmip]->GetReadPtr(cPlane), h,w,src_pitch, mips[cmip]->GetPitch(cPlane),vi.pixel_type);
    
    if (sc_next < scenechange * scenechange_factor) {
      t_th[cmip] = temp_thresh;
      t_thUV[cmip] = temp_threshUV;
      cmip++;
      mips[cmip] = curr_mips[0]->GetFrame(n+1,env); t_th[cmip] = (temp_thresh+1)/2; t_thUV[cmip] = (temp_threshUV+1)/2;
      cmip++;
      mips[cmip] = curr_mips[1]->GetFrame(n+1,env); t_th[cmip] = (temp_thresh+2)/3; t_thUV[cmip] = (temp_threshUV+2)/3;
      cmip++;
    }
  }  // end if  STRONG or SOFT

  int planecounter = 0;


  if (cmip<=1) {    // Just return the one mip used
    if (cmip) src = mips[0];    
    goto writeinfo;
  }



  int c_div;
  if (method == METHOD_STRONG) {
    c_div = 32767/(cmip-1);
  } else if (method == METHOD_SOFT) {
    c_div = 32767/cmip;
  } else if (method==METHOD_SUPERSTRONG) {
    env->MakeWritable(&src);
    c_div = 32767/(cmip-1);
  } else if (method==METHOD_SUPERSOFT) {
    env->MakeWritable(&src);
    c_div = 32767/(cmip);
  }


  while (planecounter++<4) {

    if (planecounter>1) {
      cPlane = (planecounter==2) ? PLANAR_U : PLANAR_V;
      srcp = src->GetWritePtr(cPlane);
      w = src->GetRowSize(cPlane);
      h = src->GetHeight(cPlane);
      src_pitch = src->GetPitch(cPlane);
    }

    if (!vi.IsPlanar()) {  // Skip after first loop if not planar
      planecounter=4;
    }

// Get planes and arrange them for loop.
    for (int i=0;i<cmip;i++) {
      planeP[cmip-i] = mips[i]->GetReadPtr(cPlane);
      planePitch[cmip-i] = mips[i]->GetPitch(cPlane);
      for (int o=0;o<8;o++) {
        thresholds[(cmip-i-1)*8+o] = (cPlane==PLANAR_Y) ? t_th[i] : t_thUV[i];
      }
    }

    if (supertwo) {
      // Add current frame supermip again if Super mode
      planeP[cmip+1] = mips[0]->GetReadPtr(cPlane);
      planePitch[cmip+1] = mips[0]->GetPitch(cPlane);
      for (int y=0;y<h;y++) {
        isse_blendline_soft(srcp, planeP, cmip, w, 32767/(cmip+1));
        srcp += src_pitch;
        for (int i=0 ;i<=cmip;i++) {
          planeP[i] += planePitch[i];
        }
      }
    } else {
    for (int y=0;y<h;y++) {
      if (method == METHOD_SOFT) {
        isse_accumulate_line_soft(srcp, planeP, cmip-1, w, thresholds, c_div);
      } else if (method == METHOD_STRONG) {
        isse_accumulate_line_strong(srcp, planeP, cmip-1, w, thresholds, c_div);
      } else if (method == METHOD_SUPERSOFT) {
        isse_accumulate_line_org_soft(srcp, planeP, cmip-1, w, *(int*)thresholds, c_div);
      } else if (method == METHOD_SUPERSTRONG) {
        isse_accumulate_line_org(srcp, planeP, cmip-1, w, *(int*)thresholds, c_div);
      }
      srcp += src_pitch;
      for (int i=0 ;i<cmip;i++) {
        planeP[i] += planePitch[i];
      }
    }
   }
  }

writeinfo:
  if (show && (!vi.IsRGB24())) {

    env->MakeWritable(&src);

    int line = 0;
    sprintf(buf,"Frame:%d",n);
    (vi.IsYV12()) ? DrawString(src, 0, line++, buf) : (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, buf) :DrawStringYUY2(src, 0, line++, buf);

    sprintf(buf,"Difference to Previous:%7.3f", (double)sc_prev/(scenechange_factor * ((vi.width/32)*32)*vi.height));
    (vi.IsYV12()) ? DrawString(src, 0, line++, buf) : (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, buf) :DrawStringYUY2(src, 0, line++, buf);

    sprintf(buf,"Difference to Next:%7.3f", (double)sc_next/(scenechange_factor * ((vi.width/32)*32)*vi.height));
    (vi.IsYV12()) ? DrawString(src, 0, line++, buf) : (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, buf) :DrawStringYUY2(src, 0, line++, buf);

    sprintf(buf,"Mips used:%d",(method == METHOD_SUPERSTRONG || method == METHOD_SUPERSOFT) ? supermips : cmip);
    (vi.IsYV12()) ? DrawString(src, 0, line++, buf) : (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, buf) :DrawStringYUY2(src, 0, line++, buf);

    line++;

    if (vi.IsYV12()) {
      sprintf(buf,"Luma Spatial:%d, Temporal:%d", spat_thresh, temp_thresh);
      DrawString(src, 0, line++, buf);

      sprintf(buf,"Chroma Spatial:%d, Temporal:%d", spat_threshUV, temp_threshUV );
      DrawString(src, 0, line++, buf);
    } else { 
      sprintf(buf,"Spatial:%d, Temporal:%d", spat_thresh, temp_thresh);
      (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, buf) : DrawStringYUY2(src, 0, line++, buf);
    }

    sprintf(buf,"Scenechange:%5.3f", scenechange/(double)(((vi.width/32)*32)*vi.height) );
    (vi.IsYV12()) ? DrawString(src, 0, line++, buf) : (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, buf) :DrawStringYUY2(src, 0, line++, buf);

    sprintf(buf,"Scalefactor:%5.1f%%,%5.1f%%,%5.1f%%", 100.0*scale_factor, 100.0*scale_factor*scale_factor, 100.0*scale_factor*scale_factor*scale_factor);
    (vi.IsYV12()) ? DrawString(src, 0, line++, buf) : (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, buf) :DrawStringYUY2(src, 0, line++, buf);

    if (method == METHOD_STRONG) {
      sprintf(buf,"Method: Strong");
    } else if (method == METHOD_SOFT) {
      sprintf(buf,"Method: Soft");
    } else if (method == METHOD_SUPERSOFT) {
      sprintf(buf,"Method: SuperMip Soft");
    } else if (method == METHOD_SUPERSTRONG) {
      sprintf(buf,"Method: SuperMip Strong");
    }

    (vi.IsYV12()) ? DrawString(src, 0, line++, buf) : (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, buf) :DrawStringYUY2(src, 0, line++, buf);
    if (weigh) {
      (vi.IsYV12()) ? DrawString(src, 0, line++, "Weigh: True") : (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, "Weigh: True") :DrawStringYUY2(src, 0, line++, "Weigh: True");
    } else {
      (vi.IsYV12()) ? DrawString(src, 0, line++, "Weigh: False") : (vi.IsRGB32()) ? DrawStringRGB32(src, 0, line++, "Weigh: False") :DrawStringYUY2(src, 0, line++, "Weigh: False");
    }
  }

  delete[] t_th;
  delete[] t_thUV;
  return src;
}




AVSValue __cdecl Create_MipSmooth(AVSValue args, void* user_data, IScriptEnvironment* env)
{
  AVSValue defaults[11] = {0,5,4,6,5,3.0,"supersoft","bilinear", "bilinear",0.65, true};
  const char* preset = args[11].AsString("");

  if (lstrcmpi(preset, "movieHQ") == 0) {
    defaults[1] = 4;  // Spatial
    defaults[2] = 5;  // Temporal
    defaults[3] = 5;  // Chroma spat
    defaults[4] = 7;  // Chroma temp
    defaults[5] = 2.0;  // Scenech
    defaults[6] = "supersoft";  // Method
    defaults[7] = "bilinear"; // Downsizer
    defaults[8] = "bilinear"; // Upsizer
    defaults[9] = 0.85;   // ScaleFactor
    defaults[10] = true;   // Weigh
  }

  if (lstrcmpi(preset, "movieHQ2") == 0) {
    defaults[1] = 2;  // Spatial
    defaults[2] = 2;  // Temporal
    defaults[3] = 3;  // Chroma spat
    defaults[4] = 3;  // Chroma temp
    defaults[5] = 2.0;  // Scenech
    defaults[6] = "superstrong";  // Method
    defaults[7] = "bilinear"; // Downsizer
    defaults[8] = "bilinear"; // Upsizer
    defaults[9] = 0.75;   // ScaleFactor
    defaults[10] = true;   // Weigh
  }

  if (lstrcmpi(preset, "movieLQ") == 0) {
    defaults[1] = 8;  // Spatial
    defaults[2] = 7;  // Temporal
    defaults[3] = 10;  // Chroma spat
    defaults[4] = 8;  // Chroma temp
    defaults[5] = 3.5;  // Scenech
    defaults[6] = "supersoft";  // Method
    defaults[7] = "bilinear"; // Downsizer
    defaults[8] = "bilinear"; // Upsizer
    defaults[9] = 0.70;   // ScaleFactor
    defaults[10] = true;   // Weigh
  }

  if (lstrcmpi(preset, "movieLQ2") == 0) {
    defaults[1] = 4;  // Spatial
    defaults[2] = 4;  // Temporal
    defaults[3] = 5;  // Chroma spat
    defaults[4] = 5;  // Chroma temp
    defaults[5] = 3.5;  // Scenech
    defaults[6] = "superstrong";  // Method
    defaults[7] = "bilinear"; // Downsizer
    defaults[8] = "bilinear"; // Upsizer
    defaults[9] = 0.70;   // ScaleFactor
    defaults[10] = true;   // Weigh
  }

  if (lstrcmpi(preset, "VHS") == 0) {
    defaults[1] = 9;  // Spatial
    defaults[2] = 6;  // Temporal
    defaults[3] = 10;  // Chroma spat
    defaults[4] = 9;  // Chroma temp
    defaults[5] = 5.0;  // Scenech
    defaults[6] = "superstrong";  // Method
    defaults[7] = "bilinear"; // Downsizer
    defaults[8] = "bilinear"; // Upsizer
    defaults[9] = 0.6;   // ScaleFactor
    defaults[10] = true;   // Weigh
  }

  if (lstrcmpi(preset, "animeHQ") == 0) {
    defaults[1] = 6;  // Spatial
    defaults[2] = 8;  // Temporal
    defaults[3] = 6;  // Chroma spat
    defaults[4] = 8;  // Chroma temp
    defaults[5] = 5.0;  // Scenech
    defaults[6] = "supersoft";  // Method
    defaults[7] = "bicubic"; // Downsizer
    defaults[8] = "bicubic"; // Upsizer
    defaults[9] = 0.60;   // ScaleFactor
    defaults[10] = true;   // Weigh
  }

  if (lstrcmpi(preset, "animeLQ") == 0) {
    defaults[1] = 5;  // Spatial
    defaults[2] = 5;  // Temporal
    defaults[3] = 5;  // Chroma spat
    defaults[4] = 7;  // Chroma temp
    defaults[5] = 5.5;  // Scenech
    defaults[6] = "superstrong";  // Method
    defaults[7] = "bilinear"; // Downsizer
    defaults[8] = "bicubic"; // Upsizer
    defaults[9] = 0.5;   // ScaleFactor
    defaults[10] = true;   // Weigh
  }

  if (lstrcmpi(preset, "rainbow") == 0) {
    defaults[1] = 0;  // Spatial
    defaults[2] = 0;  // Temporal
    defaults[3] = 255;  // Chroma spat
    defaults[4] = 255;  // Chroma temp
    defaults[5] = 2.0;  // Scenech
    defaults[6] = "superstrong";  // Method
    defaults[7] = "bilinear"; // Downsizer
    defaults[8] = "bilinear"; // Upsizer
    defaults[9] = 0.65;   // ScaleFactor
    defaults[10] = true;   // Weigh
  }
  
  bool IsCustom = (lstrcmpi(preset, "custom") == 0);

  if (IsCustom) {
    int read_error = 0;
    HKEY AvisynthKey;
    DWORD ReadDWord;
    DWORD DWordSize=sizeof(DWORD);
    LPBYTE ReadString;
    
    if (!RegOpenKeyEx(RegRootKey, RegAvisynthKey, 0, KEY_READ, &AvisynthKey)) {
      if (!RegQueryValueEx(AvisynthKey, "MipSmoothPreset", 0, 0, 0, &ReadDWord)) { // We now assume we can read without error.

        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothSpYThreshold", 0, 0, (LPBYTE)&ReadDWord, &DWordSize);
        defaults[1] = AVSValue((int)ReadDWord);
        
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothSpUVThreshold", 0, 0, (LPBYTE)&ReadDWord, &DWordSize);
        defaults[3] = AVSValue((int)ReadDWord);
        
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothTeYThreshold", 0, 0, (LPBYTE)&ReadDWord, &DWordSize);
        defaults[2] = AVSValue((int)ReadDWord);
        
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothTeUVThreshold", 0, 0, (LPBYTE)&ReadDWord, &DWordSize);
        defaults[4] = AVSValue((int)ReadDWord);
        
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothScenechange", 0, 0, (LPBYTE)&ReadDWord,&DWordSize);
        defaults[5] = AVSValue((double)ReadDWord/10000.0);
        
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothResizeFactor", 0, 0, (LPBYTE)&ReadDWord, &DWordSize);
        defaults[9] = AVSValue((double)ReadDWord/10000.0);
        
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothMethod", 0, 0, 0, &ReadDWord);
        ReadString = new unsigned char[ReadDWord];
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothMethod", 0, 0, ReadString, &ReadDWord);
        defaults[6] = AVSValue((const char*)ReadString);
        
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothUpsizer", 0, 0, 0, &ReadDWord);
        LPBYTE ReadString2 = new unsigned char[ReadDWord];
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothUpsizer", 0, 0, ReadString2, &ReadDWord);
        defaults[8] = AVSValue((const char*)ReadString2);
        
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothDownsizer", 0, 0, 0, &ReadDWord);
        LPBYTE ReadString3 = new unsigned char[ReadDWord];

        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothDownsizer", 0, 0, ReadString3, &ReadDWord);
        defaults[7] = AVSValue((const char*)ReadString3);
        
        read_error += RegQueryValueEx(AvisynthKey, "MipSmoothWeigh", 0, 0, (LPBYTE)&ReadDWord, &DWordSize);
        defaults[10] = AVSValue(!!ReadDWord);

      } else env->ThrowError("MipSmooth: No stored preset was found!");
    } else env->ThrowError("MipSmooth: Could not find the AviSynth registry entries. Reinstall AviSynth!");
    
    if (read_error) 
      env->ThrowError("MipSmooth: No custom preset could be found, or problems reading entries!");        
  }

  bool StoreCustom = args[12].AsBool(false);
  
  if (StoreCustom) {
    HKEY AvisynthKey;
    DWORD WriteWord;
    const char* WriteString;

    // Construct settings
    AVSValue settings[11] = {0,
      args[1].AsInt(defaults[1].AsInt()), // Spatial
      args[2].AsInt(defaults[2].AsInt()),  // Temporal
      args[3].AsInt(args[1].AsInt(defaults[3].AsInt())), // Spatial chroma
      args[4].AsInt(args[2].AsInt(defaults[4].AsInt())),  // Temporal chroma
      args[5].AsFloat(defaults[5].AsFloat()), // Scenechange
      args[6].AsString(defaults[6].AsString()),  // method
      args[7].AsString(defaults[7].AsString()),  // downsizer
      args[8].AsString(defaults[8].AsString()),  // Upsizer
      args[9].AsFloat(defaults[9].AsFloat()),  // Scalefactor
      args[10].AsFloat(defaults[10].AsBool()),  // Weigh
    };
    
    if (!RegOpenKeyEx(RegRootKey, RegAvisynthKey, 0, KEY_SET_VALUE, &AvisynthKey)) {
      
      int write_error = 0;
      
      WriteWord = 0;
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothPreset", 0, REG_DWORD,(LPBYTE)&WriteWord, sizeof(DWORD));
      
      if (write_error)
        env->ThrowError("MipSmooth: Could not write custom presets to registry.");
      
      WriteWord = settings[1].AsInt();
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothSpYThreshold", 0, REG_DWORD,(LPBYTE)&WriteWord, sizeof(DWORD));
      
      WriteWord = settings[2].AsInt();
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothTeYThreshold", 0, REG_DWORD,(LPBYTE)&WriteWord, sizeof(DWORD));
      
      WriteWord = settings[3].AsInt();
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothSpUVThreshold", 0, REG_DWORD,(LPBYTE)&WriteWord, sizeof(DWORD));
      
      WriteWord = settings[4].AsInt();
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothTeUVThreshold", 0, REG_DWORD,(LPBYTE)&WriteWord, sizeof(DWORD));
      
      WriteWord = (int)(settings[5].AsFloat()*10000.0);
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothScenechange", 0, REG_DWORD,(LPBYTE)&WriteWord, sizeof(DWORD));
      
      WriteWord = (int)(settings[9].AsFloat()*10000.0);
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothResizefactor", 0, REG_DWORD,(LPBYTE)&WriteWord, sizeof(DWORD));
      
      WriteString = settings[6].AsString();
      WriteWord = strlen(WriteString);
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothMethod", 0, REG_SZ, (LPBYTE)WriteString, WriteWord);
      
      WriteString = settings[7].AsString();
      WriteWord = strlen(WriteString);
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothDownsizer", 0, REG_SZ, (LPBYTE)WriteString, WriteWord);
      
      WriteString = settings[8].AsString();
      WriteWord = strlen(WriteString);
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothUpsizer", 0, REG_SZ, (LPBYTE)WriteString, WriteWord);

      WriteWord = (settings[10].AsBool()) ? 1 : 0;
      write_error += RegSetValueEx(AvisynthKey, "MipSmoothWeigh", 0, REG_DWORD,(LPBYTE)&WriteWord, sizeof(DWORD));
      
      if (write_error)
        env->ThrowError("MipSmooth: An error ocurred while saving custom preset. Check you have enough security rights.");
      RegCloseKey(AvisynthKey);
    } else env->ThrowError("MipSmooth: Unable to store presets - AviSynth registry entries not found, or no write permission!");

  }

  MipSmooth *m = new MipSmooth(args[0].AsClip(),
      args[1].AsInt(defaults[1].AsInt()), // Spatial
      args[2].AsInt(defaults[2].AsInt()),  // Temporal
      args[3].AsInt(args[1].AsInt(defaults[3].AsInt())), // Spatial chroma
      args[4].AsInt(args[2].AsInt(defaults[4].AsInt())),  // Temporal chroma
      args[5].AsFloat(defaults[5].AsFloat()), // Scenechange
      args[6].AsString(defaults[6].AsString()),  // method
      args[7].AsString(defaults[7].AsString()),  // downsizer
      args[8].AsString(defaults[8].AsString()),  // Upsizer
      args[9].AsFloat(defaults[9].AsFloat()),  // Scalefactor
      args[13].AsBool(false),                 // show
      args[10].AsBool(defaults[10].AsBool()),     // weigh 
      env);

  return m;

}


extern "C" __declspec(dllexport) const char* __stdcall AvisynthPluginInit2(IScriptEnvironment* env)
{
    env->AddFunction("MipSmooth", "c[spatial]i[temporal]i[spatial_chroma]i[temporal_chroma]i[scenechange]f[method]s[downsizer]s[upsizer]s[scalefactor]f[weigh]b[preset]s[storecustom]b[show]b", Create_MipSmooth, 0);
    return 0;
}

